"""How many different ways can you climb a staircase of `n` steps?

You can climb 1, 2, or 3 steps at a time.

For one step, we could do this one way: 1

    >>> steps(1)
    1

For two: 11 2

    >>> steps(2)
    2

For three: 111 12 21 3

    >>> steps(3)
    4

For four: 1111 121 211 112 22 13 31

    >>> steps(4)
    7

For five: 11111 2111 1211 1121 1112 122 121 221 113 131 311 23 32

    >>> steps(5)
    13

For six steps: 111111 21111 12111 11211 11121 11112 2211 2121 2112
    1212 1122 1221 3111 1311 1131 1113 321 312 213 231 132 123 222 33

    >>> steps(6)
    24
"""


def steps(n):
    """How many different ways can you climb a staircase of `n` steps?

    You can climb 1, 2, or 3 steps at a time.
    """

    # START SOLUTION

    # Base case:
    #
    # It might seem strange to think there's a way to take zero
    # staircase-steps --- but this is really a saying "this path of recursion"
    # is successful, since it ended up using exactly the right number
    # of staircase-steps

    if n == 0:
        return 1

    # If we overshoot, this was not a successful path
    if n < 1:
        return 0

    # The way to take n number of staircase-steps is the way to take
    # n - 1 staircase-steps, n - 2 staircase-steps, and n - 3 staircase-steps,
    # all added up (since we can always take 1, 2, or 3 more staircase-steps
    # afterwards)

    return steps(n - 1) + steps(n - 2) + steps(n - 3)


def steps_cache(n):
    """How many different ways can you climb a staircase of `num` steps?

    You can climb 1, 2, or 3 steps at a time.

    In this version, we're caching previous work, so we can do much
    longer calculations.

    For 100 steps: 180,396,380,815,100,901,214,157,639 different ways!

        >>> steps_cache(100)
        180396380815100901214157639

    """

    # Cache previous work, so we don't keep having to re-calculate
    # answers we've already derived. This changes the runtime from
    # O(3**n) to the much, much better O(3n)
    #
    # ``cache`` will hold the results of taking each number of staircase-steps

    cache = [None] * (n + 1)

    def _steps(n):
        """How many ways are there to take `n` steps?"""

        # Base case:
        #
        # It might seem strange to think there's a way to take zero
        # staircase-steps --- but this is really a saying "this path of recursion"
        # is successful, since it ended up using exactly the right number
        # of staircase-steps
        if n == 0:
            return 1

        # If we overshoot, this was not a successful path
        if n < 1:
            return 0

        # If we haven't already cached this work, do so
        if cache[n] is None:
            cache[n] = _steps(n - 1) + _steps(n - 2) + _steps(n - 3)

        return cache[n]

    return _steps(n)

# END SOLUTION


if __name__ == '__main__':
    import doctest
    if doctest.testmod().failed == 0:
        print("\n*** ALL TEST PASSED! YOU'RE A STAIRMASTER!\n")
